@@ -4,7 +4,7 @@ use anyhow::{Context, Result};
 use filetime::FileTime;
 use std::env;
 use std::ffi::{OsStr, OsString};
-use std::fs::{self, File, OpenOptions};
+use std::fs::{self, File, Metadata, OpenOptions};
 use std::io;
 use std::io::prelude::*;
 use std::iter;
@@ -55,6 +55,8 @@ pub fn dylib_path_envvar() -> &'static str {
         // penalty starting in 10.13. Cargo's testsuite ran more than twice as
         // slow with it on CI.
         "DYLD_FALLBACK_LIBRARY_PATH"
+    } else if cfg!(target_os = "aix") {
+        "LIBPATH"
     } else {
         "LD_LIBRARY_PATH"
     }
@@ -92,11 +94,18 @@ pub fn normalize_path(path: &Path) -> PathBuf {
         match component {
             Component::Prefix(..) => unreachable!(),
             Component::RootDir => {
-                ret.push(component.as_os_str());
+                ret.push(Component::RootDir);
             }
             Component::CurDir => {}
             Component::ParentDir => {
-                ret.pop();
+                if ret.ends_with(Component::ParentDir) {
+                    ret.push(Component::ParentDir);
+                } else {
+                    let popped = ret.pop();
+                    if !popped && !ret.has_root() {
+                        ret.push(Component::ParentDir);
+                    }
+                }
             }
             Component::Normal(c) => {
                 ret.push(c);
@@ -134,6 +143,24 @@ pub fn resolve_executable(exec: &Path) -> Result<PathBuf> {
     }
 }
 
+/// Returns metadata for a file (follows symlinks).
+///
+/// Equivalent to [`std::fs::metadata`] with better error messages.
+pub fn metadata<P: AsRef<Path>>(path: P) -> Result<Metadata> {
+    let path = path.as_ref();
+    std::fs::metadata(path)
+        .with_context(|| format!("failed to load metadata for path `{}`", path.display()))
+}
+
+/// Returns metadata for a file without following symlinks.
+///
+/// Equivalent to [`std::fs::metadata`] with better error messages.
+pub fn symlink_metadata<P: AsRef<Path>>(path: P) -> Result<Metadata> {
+    let path = path.as_ref();
+    std::fs::symlink_metadata(path)
+        .with_context(|| format!("failed to load metadata for path `{}`", path.display()))
+}
+
 /// Reads a file to a string.
 ///
 /// Equivalent to [`std::fs::read_to_string`] with better error messages.
@@ -160,6 +187,43 @@ pub fn write<P: AsRef<Path>, C: AsRef<[u8]>>(path: P, contents: C) -> Result<()>
         .with_context(|| format!("failed to write `{}`", path.display()))
 }
 
+/// Writes a file to disk atomically.
+///
+/// This uses `tempfile::persist` to accomplish atomic writes.
+pub fn write_atomic<P: AsRef<Path>, C: AsRef<[u8]>>(path: P, contents: C) -> Result<()> {
+    let path = path.as_ref();
+
+    // On unix platforms, get the permissions of the original file. Copy only the user/group/other
+    // read/write/execute permission bits. The tempfile lib defaults to an initial mode of 0o600,
+    // and we'll set the proper permissions after creating the file.
+    #[cfg(unix)]
+    let perms = path.metadata().ok().map(|meta| {
+        use std::os::unix::fs::PermissionsExt;
+
+        // these constants are u16 on macOS
+        let mask = u32::from(libc::S_IRWXU | libc::S_IRWXG | libc::S_IRWXO);
+        let mode = meta.permissions().mode() & mask;
+
+        std::fs::Permissions::from_mode(mode)
+    });
+
+    let mut tmp = TempFileBuilder::new()
+        .prefix(path.file_name().unwrap())
+        .tempfile_in(path.parent().unwrap())?;
+    tmp.write_all(contents.as_ref())?;
+
+    // On unix platforms, set the permissions on the newly created file. We can use fchmod (called
+    // by the std lib; subject to change) which ignores the umask so that the new file has the same
+    // permissions as the old file.
+    #[cfg(unix)]
+    if let Some(perms) = perms {
+        tmp.as_file().set_permissions(perms)?;
+    }
+
+    tmp.persist(path)?;
+    Ok(())
+}
+
 /// Equivalent to [`write()`], but does not write anything if the file contents
 /// are identical to the given contents.
 pub fn write_if_changed<P: AsRef<Path>, C: AsRef<[u8]>>(path: P, contents: C) -> Result<()> {
@@ -214,16 +278,14 @@ pub fn open<P: AsRef<Path>>(path: P) -> Result<File> {
 
 /// Returns the last modification time of a file.
 pub fn mtime(path: &Path) -> Result<FileTime> {
-    let meta =
-        fs::metadata(path).with_context(|| format!("failed to stat `{}`", path.display()))?;
+    let meta = metadata(path)?;
     Ok(FileTime::from_last_modification_time(&meta))
 }
 
 /// Returns the maximum mtime of the given path, recursing into
 /// subdirectories, and following symlinks.
 pub fn mtime_recursive(path: &Path) -> Result<FileTime> {
-    let meta =
-        fs::metadata(path).with_context(|| format!("failed to stat `{}`", path.display()))?;
+    let meta = metadata(path)?;
     if !meta.is_dir() {
         return Ok(FileTime::from_last_modification_time(&meta));
     }
@@ -235,7 +297,7 @@ pub fn mtime_recursive(path: &Path) -> Result<FileTime> {
             Err(e) => {
                 // Ignore errors while walking. If Cargo can't access it, the
                 // build script probably can't access it, either.
-                log::debug!("failed to determine mtime while walking directory: {}", e);
+                tracing::debug!("failed to determine mtime while walking directory: {}", e);
                 None
             }
         })
@@ -250,7 +312,7 @@ pub fn mtime_recursive(path: &Path) -> Result<FileTime> {
                         // I'm not sure when this is really possible (maybe a
                         // race with unlinking?). Regardless, if Cargo can't
                         // read it, the build script probably can't either.
-                        log::debug!(
+                        tracing::debug!(
                             "failed to determine mtime while fetching symlink metadata of {}: {}",
                             e.path().display(),
                             err
@@ -269,7 +331,7 @@ pub fn mtime_recursive(path: &Path) -> Result<FileTime> {
                         // Can't access the symlink target. If Cargo can't
                         // access it, the build script probably can't access
                         // it either.
-                        log::debug!(
+                        tracing::debug!(
                             "failed to determine mtime of symlink target for {}: {}",
                             e.path().display(),
                             err
@@ -284,7 +346,7 @@ pub fn mtime_recursive(path: &Path) -> Result<FileTime> {
                         // I'm not sure when this is really possible (maybe a
                         // race with unlinking?). Regardless, if Cargo can't
                         // read it, the build script probably can't either.
-                        log::debug!(
+                        tracing::debug!(
                             "failed to determine mtime while fetching metadata of {}: {}",
                             e.path().display(),
                             err
@@ -312,7 +374,7 @@ pub fn set_invocation_time(path: &Path) -> Result<FileTime> {
         "This file has an mtime of when this was started.",
     )?;
     let ft = mtime(&timestamp)?;
-    log::debug!("invocation time for {:?} is {}", path, ft);
+    tracing::debug!("invocation time for {:?} is {}", path, ft);
     Ok(ft)
 }
 
@@ -411,18 +473,26 @@ fn _create_dir_all(p: &Path) -> Result<()> {
     Ok(())
 }
 
-/// Recursively remove all files and directories at the given directory.
+/// Equivalent to [`std::fs::remove_dir_all`] with better error messages.
 ///
 /// This does *not* follow symlinks.
 pub fn remove_dir_all<P: AsRef<Path>>(p: P) -> Result<()> {
-    _remove_dir_all(p.as_ref())
+    _remove_dir_all(p.as_ref()).or_else(|prev_err| {
+        // `std::fs::remove_dir_all` is highly specialized for different platforms
+        // and may be more reliable than a simple walk. We try the walk first in
+        // order to report more detailed errors.
+        fs::remove_dir_all(p.as_ref()).with_context(|| {
+            format!(
+                "{:?}\n\nError: failed to remove directory `{}`",
+                prev_err,
+                p.as_ref().display(),
+            )
+        })
+    })
 }
 
 fn _remove_dir_all(p: &Path) -> Result<()> {
-    if p.symlink_metadata()
-        .with_context(|| format!("could not get metadata for `{}` to remove", p.display()))?
-        .is_symlink()
-    {
+    if symlink_metadata(p)?.is_symlink() {
         return remove_file(p);
     }
     let entries = p
@@ -454,25 +524,57 @@ fn _remove_dir(p: &Path) -> Result<()> {
 ///
 /// If the file is readonly, this will attempt to change the permissions to
 /// force the file to be deleted.
+/// On Windows, if the file is a symlink to a directory, this will attempt to remove
+/// the symlink itself.
 pub fn remove_file<P: AsRef<Path>>(p: P) -> Result<()> {
     _remove_file(p.as_ref())
 }
 
 fn _remove_file(p: &Path) -> Result<()> {
-    let mut err = match fs::remove_file(p) {
-        Ok(()) => return Ok(()),
-        Err(e) => e,
-    };
-
-    if err.kind() == io::ErrorKind::PermissionDenied && set_not_readonly(p).unwrap_or(false) {
-        match fs::remove_file(p) {
-            Ok(()) => return Ok(()),
-            Err(e) => err = e,
+    // For Windows, we need to check if the file is a symlink to a directory
+    // and remove the symlink itself by calling `remove_dir` instead of
+    // `remove_file`.
+    #[cfg(target_os = "windows")]
+    {
+        use std::os::windows::fs::FileTypeExt;
+        let metadata = symlink_metadata(p)?;
+        let file_type = metadata.file_type();
+        if file_type.is_symlink_dir() {
+            return remove_symlink_dir_with_permission_check(p);
         }
     }
 
-    Err(err).with_context(|| format!("failed to remove file `{}`", p.display()))?;
-    Ok(())
+    remove_file_with_permission_check(p)
+}
+
+#[cfg(target_os = "windows")]
+fn remove_symlink_dir_with_permission_check(p: &Path) -> Result<()> {
+    remove_with_permission_check(fs::remove_dir, p)
+        .with_context(|| format!("failed to remove symlink dir `{}`", p.display()))
+}
+
+fn remove_file_with_permission_check(p: &Path) -> Result<()> {
+    remove_with_permission_check(fs::remove_file, p)
+        .with_context(|| format!("failed to remove file `{}`", p.display()))
+}
+
+fn remove_with_permission_check<F, P>(remove_func: F, p: P) -> io::Result<()>
+where
+    F: Fn(P) -> io::Result<()>,
+    P: AsRef<Path> + Clone,
+{
+    match remove_func(p.clone()) {
+        Ok(()) => Ok(()),
+        Err(e) => {
+            if e.kind() == io::ErrorKind::PermissionDenied
+                && set_not_readonly(p.as_ref()).unwrap_or(false)
+            {
+                remove_func(p)
+            } else {
+                Err(e)
+            }
+        }
+    }
 }
 
 fn set_not_readonly(p: &Path) -> io::Result<bool> {
@@ -495,7 +597,7 @@ pub fn link_or_copy(src: impl AsRef<Path>, dst: impl AsRef<Path>) -> Result<()>
 }
 
 fn _link_or_copy(src: &Path, dst: &Path) -> Result<()> {
-    log::debug!("linking {} to {}", src.display(), dst.display());
+    tracing::debug!("linking {} to {}", src.display(), dst.display());
     if same_file::is_same_file(src, dst).unwrap_or(false) {
         return Ok(());
     }
@@ -527,34 +629,42 @@ fn _link_or_copy(src: &Path, dst: &Path) -> Result<()> {
             src
         };
         symlink(src, dst)
-    } else if env::var_os("__CARGO_COPY_DONT_LINK_DO_NOT_USE_THIS").is_some() {
-        // This is a work-around for a bug in macOS 10.15. When running on
-        // APFS, there seems to be a strange race condition with
-        // Gatekeeper where it will forcefully kill a process launched via
-        // `cargo run` with SIGKILL. Copying seems to avoid the problem.
-        // This shouldn't affect anyone except Cargo's test suite because
-        // it is very rare, and only seems to happen under heavy load and
-        // rapidly creating lots of executables and running them.
-        // See https://github.com/rust-lang/cargo/issues/7821 for the
-        // gory details.
-        fs::copy(src, dst).map(|_| ())
     } else {
         if cfg!(target_os = "macos") {
-            // This is a work-around for a bug on macos. There seems to be a race condition
-            // with APFS when hard-linking binaries. Gatekeeper does not have signing or
-            // hash information stored in kernel when running the process. Therefore killing it.
-            // This problem does not appear when copying files as kernel has time to process it.
-            // Note that: fs::copy on macos is using CopyOnWrite (syscall fclonefileat) which should be
-            // as fast as hardlinking.
-            // See https://github.com/rust-lang/cargo/issues/10060 for the details
-            fs::copy(src, dst).map(|_| ())
+            // There seems to be a race condition with APFS when hard-linking
+            // binaries. Gatekeeper does not have signing or hash information
+            // stored in kernel when running the process. Therefore killing it.
+            // This problem does not appear when copying files as kernel has
+            // time to process it. Note that: fs::copy on macos is using
+            // CopyOnWrite (syscall fclonefileat) which should be as fast as
+            // hardlinking. See these issues for the details:
+            //
+            // * https://github.com/rust-lang/cargo/issues/7821
+            // * https://github.com/rust-lang/cargo/issues/10060
+            fs::copy(src, dst).map_or_else(
+                |e| {
+                    if e.raw_os_error()
+                        .map_or(false, |os_err| os_err == 35 /* libc::EAGAIN */)
+                    {
+                        tracing::info!("copy failed {e:?}. falling back to fs::hard_link");
+
+                        // Working around an issue copying too fast with zfs (probably related to
+                        // https://github.com/openzfsonosx/zfs/issues/809)
+                        // See https://github.com/rust-lang/cargo/issues/13838
+                        fs::hard_link(src, dst)
+                    } else {
+                        Err(e)
+                    }
+                },
+                |_| Ok(()),
+            )
         } else {
             fs::hard_link(src, dst)
         }
     };
     link_result
         .or_else(|err| {
-            log::debug!("link failed {}. falling back to fs::copy", err);
+            tracing::debug!("link failed {}. falling back to fs::copy", err);
             fs::copy(src, dst).map(|_| ())
         })
         .with_context(|| {
@@ -585,8 +695,8 @@ pub fn copy<P: AsRef<Path>, Q: AsRef<Path>>(from: P, to: Q) -> Result<u64> {
 pub fn set_file_time_no_err<P: AsRef<Path>>(path: P, time: FileTime) {
     let path = path.as_ref();
     match filetime::set_file_times(path, time, time) {
-        Ok(()) => log::debug!("set file mtime {} to {}", path.display(), time),
-        Err(e) => log::warn!(
+        Ok(()) => tracing::debug!("set file mtime {} to {}", path.display(), time),
+        Err(e) => tracing::warn!(
             "could not set mtime of {} to {}: {:?}",
             path.display(),
             time,
@@ -600,15 +710,15 @@ pub fn set_file_time_no_err<P: AsRef<Path>>(path: P, time: FileTime) {
 /// This canonicalizes both paths before stripping. This is useful if the
 /// paths are obtained in different ways, and one or the other may or may not
 /// have been normalized in some way.
-pub fn strip_prefix_canonical<P: AsRef<Path>>(
-    path: P,
-    base: P,
+pub fn strip_prefix_canonical(
+    path: impl AsRef<Path>,
+    base: impl AsRef<Path>,
 ) -> Result<PathBuf, std::path::StripPrefixError> {
     // Not all filesystems support canonicalize. Just ignore if it doesn't work.
     let safe_canonicalize = |path: &Path| match path.canonicalize() {
         Ok(p) => p,
         Err(e) => {
-            log::warn!("cannot canonicalize {:?}: {:?}", path, e);
+            tracing::warn!("cannot canonicalize {:?}: {:?}", path, e);
             path.to_path_buf()
         }
     };
@@ -655,7 +765,8 @@ pub fn create_dir_all_excluded_from_backups_atomic(p: impl AsRef<Path>) -> Resul
     // we can infer from it's another cargo process doing work.
     if let Err(e) = fs::rename(tempdir.path(), path) {
         if !path.exists() {
-            return Err(anyhow::Error::from(e));
+            return Err(anyhow::Error::from(e))
+                .with_context(|| format!("failed to create directory `{}`", path.display()));
         }
     }
     Ok(())
@@ -679,14 +790,17 @@ pub fn exclude_from_backups_and_indexing(p: impl AsRef<Path>) {
 /// * CACHEDIR.TAG files supported by various tools in a platform-independent way
 fn exclude_from_backups(path: &Path) {
     exclude_from_time_machine(path);
-    let _ = std::fs::write(
-        path.join("CACHEDIR.TAG"),
-        "Signature: 8a477f597d28d172789f06886806bc55
+    let file = path.join("CACHEDIR.TAG");
+    if !file.exists() {
+        let _ = std::fs::write(
+            file,
+            "Signature: 8a477f597d28d172789f06886806bc55
 # This file is a cache directory tag created by cargo.
 # For information about cache directory tags see https://bford.info/cachedir/
 ",
-    );
-    // Similarly to exclude_from_time_machine() we ignore errors here as it's an optional feature.
+        );
+        // Similarly to exclude_from_time_machine() we ignore errors here as it's an optional feature.
+    }
 }
 
 /// Marks the directory as excluded from content indexing.
@@ -749,6 +863,89 @@ fn exclude_from_time_machine(path: &Path) {
 #[cfg(test)]
 mod tests {
     use super::join_paths;
+    use super::normalize_path;
+    use super::write;
+    use super::write_atomic;
+
+    #[test]
+    fn test_normalize_path() {
+        let cases = &[
+            ("", ""),
+            (".", ""),
+            (".////./.", ""),
+            ("/", "/"),
+            ("/..", "/"),
+            ("/foo/bar", "/foo/bar"),
+            ("/foo/bar/", "/foo/bar"),
+            ("/foo/bar/./././///", "/foo/bar"),
+            ("/foo/bar/..", "/foo"),
+            ("/foo/bar/../..", "/"),
+            ("/foo/bar/../../..", "/"),
+            ("foo/bar", "foo/bar"),
+            ("foo/bar/", "foo/bar"),
+            ("foo/bar/./././///", "foo/bar"),
+            ("foo/bar/..", "foo"),
+            ("foo/bar/../..", ""),
+            ("foo/bar/../../..", ".."),
+            ("../../foo/bar", "../../foo/bar"),
+            ("../../foo/bar/", "../../foo/bar"),
+            ("../../foo/bar/./././///", "../../foo/bar"),
+            ("../../foo/bar/..", "../../foo"),
+            ("../../foo/bar/../..", "../.."),
+            ("../../foo/bar/../../..", "../../.."),
+        ];
+        for (input, expected) in cases {
+            let actual = normalize_path(std::path::Path::new(input));
+            assert_eq!(actual, std::path::Path::new(expected), "input: {input}");
+        }
+    }
+
+    #[test]
+    fn write_works() {
+        let original_contents = "[dependencies]\nfoo = 0.1.0";
+
+        let tmpdir = tempfile::tempdir().unwrap();
+        let path = tmpdir.path().join("Cargo.toml");
+        write(&path, original_contents).unwrap();
+        let contents = std::fs::read_to_string(&path).unwrap();
+        assert_eq!(contents, original_contents);
+    }
+    #[test]
+    fn write_atomic_works() {
+        let original_contents = "[dependencies]\nfoo = 0.1.0";
+
+        let tmpdir = tempfile::tempdir().unwrap();
+        let path = tmpdir.path().join("Cargo.toml");
+        write_atomic(&path, original_contents).unwrap();
+        let contents = std::fs::read_to_string(&path).unwrap();
+        assert_eq!(contents, original_contents);
+    }
+
+    #[test]
+    #[cfg(unix)]
+    fn write_atomic_permissions() {
+        use std::os::unix::fs::PermissionsExt;
+
+        let original_perms = std::fs::Permissions::from_mode(u32::from(
+            libc::S_IRWXU | libc::S_IRGRP | libc::S_IWGRP | libc::S_IROTH,
+        ));
+
+        let tmp = tempfile::Builder::new().tempfile().unwrap();
+
+        // need to set the permissions after creating the file to avoid umask
+        tmp.as_file()
+            .set_permissions(original_perms.clone())
+            .unwrap();
+
+        // after this call, the file at `tmp.path()` will not be the same as the file held by `tmp`
+        write_atomic(tmp.path(), "new").unwrap();
+        assert_eq!(std::fs::read_to_string(tmp.path()).unwrap(), "new");
+
+        let new_perms = std::fs::metadata(tmp.path()).unwrap().permissions();
+
+        let mask = u32::from(libc::S_IRWXU | libc::S_IRWXG | libc::S_IRWXO);
+        assert_eq!(original_perms.mode(), new_perms.mode() & mask);
+    }
 
     #[test]
     fn join_paths_lists_paths_on_error() {
@@ -785,4 +982,50 @@ mod tests {
             );
         }
     }
+
+    #[test]
+    #[cfg(windows)]
+    fn test_remove_symlink_dir() {
+        use super::*;
+        use std::fs;
+        use std::os::windows::fs::symlink_dir;
+
+        let tmpdir = tempfile::tempdir().unwrap();
+        let dir_path = tmpdir.path().join("testdir");
+        let symlink_path = tmpdir.path().join("symlink");
+
+        fs::create_dir(&dir_path).unwrap();
+
+        symlink_dir(&dir_path, &symlink_path).expect("failed to create symlink");
+
+        assert!(symlink_path.exists());
+
+        assert!(remove_file(symlink_path.clone()).is_ok());
+
+        assert!(!symlink_path.exists());
+        assert!(dir_path.exists());
+    }
+
+    #[test]
+    #[cfg(windows)]
+    fn test_remove_symlink_file() {
+        use super::*;
+        use std::fs;
+        use std::os::windows::fs::symlink_file;
+
+        let tmpdir = tempfile::tempdir().unwrap();
+        let file_path = tmpdir.path().join("testfile");
+        let symlink_path = tmpdir.path().join("symlink");
+
+        fs::write(&file_path, b"test").unwrap();
+
+        symlink_file(&file_path, &symlink_path).expect("failed to create symlink");
+
+        assert!(symlink_path.exists());
+
+        assert!(remove_file(symlink_path.clone()).is_ok());
+
+        assert!(!symlink_path.exists());
+        assert!(file_path.exists());
+    }
 }
