@@ -91,9 +91,11 @@ enum class ErrorType {
 extern "C" {
 /// Print additional debug info relevant to performance
 extern llvm::cl::opt<bool> EnzymePrintPerf;
+extern llvm::cl::opt<bool> EnzymeNonPower2Cache;
 extern llvm::cl::opt<bool> EnzymeStrongZero;
 extern llvm::cl::opt<bool> EnzymeBlasCopy;
 extern llvm::cl::opt<bool> EnzymeLapackCopy;
+extern llvm::cl::opt<bool> EnzymeJuliaAddrLoad;
 extern LLVMValueRef (*CustomErrorHandler)(const char *, LLVMValueRef, ErrorType,
                                           const void *, LLVMValueRef,
                                           LLVMBuilderRef);
@@ -677,7 +679,7 @@ struct BlasInfo {
   std::string function;
   bool is64;
 
-  llvm::Type *fpType(llvm::LLVMContext &ctx) const;
+  llvm::Type *fpType(llvm::LLVMContext &ctx, bool to_scalar = false) const;
   llvm::IntegerType *intType(llvm::LLVMContext &ctx) const;
 };
 
@@ -687,6 +689,9 @@ std::optional<BlasInfo> extractBLAS(llvm::StringRef in);
 llvm::Optional<BlasInfo> extractBLAS(llvm::StringRef in);
 #endif
 
+std::vector<std::tuple<llvm::Type *, size_t, size_t>>
+parseTrueType(const llvm::MDNode *, DerivativeMode, bool const_src);
+
 /// Create function for type that performs the derivative memcpy on floating
 /// point memory
 llvm::Function *getOrInsertDifferentialFloatMemcpy(
@@ -1156,7 +1161,7 @@ static inline llvm::StringRef getFuncName(llvm::Function *called) {
     return called->getName();
 }
 
-template <typename T> static inline llvm::StringRef getFuncNameFromCall(T *op) {
+static inline llvm::StringRef getFuncNameFromCall(const llvm::CallBase *op) {
   auto AttrList =
       op->getAttributes().getAttributes(llvm::AttributeList::FunctionIndex);
   if (AttrList.hasAttribute("enzyme_math"))
@@ -1170,11 +1175,39 @@ template <typename T> static inline llvm::StringRef getFuncNameFromCall(T *op) {
   return "";
 }
 
-template <typename T>
+static inline bool hasNoCache(llvm::Value *op) {
+  using namespace llvm;
+  if (auto CB = dyn_cast<CallBase>(op)) {
+    if (auto called = getFunctionFromCall(CB)) {
+      if (called->hasFnAttribute("enzyme_nocache"))
+        return true;
+    }
+  }
+  if (auto I = dyn_cast<Instruction>(op))
+    if (hasMetadata(I, "enzyme_nocache"))
+      return true;
+
+  if (EnzymeJuliaAddrLoad) {
+    if (auto PT = dyn_cast<PointerType>(op->getType())) {
+      if (PT->getAddressSpace() == 11 || PT->getAddressSpace() == 13) {
+        if (isa<CastInst>(op) || isa<GetElementPtrInst>(op))
+          return true;
+      }
+    }
+  }
+  if (auto IT = dyn_cast<IntegerType>(op->getType()))
+    if (!isPowerOf2_64(IT->getBitWidth()) && !EnzymeNonPower2Cache)
+      return true;
+
+  return false;
+}
+
 #if LLVM_VERSION_MAJOR >= 16
-static inline std::optional<size_t> getAllocationIndexFromCall(T *op)
+static inline std::optional<size_t>
+getAllocationIndexFromCall(const llvm::CallBase *op)
 #else
-static inline llvm::Optional<size_t> getAllocationIndexFromCall(T *op)
+static inline llvm::Optional<size_t>
+getAllocationIndexFromCall(const llvm::CallBase *op)
 #endif
 {
   auto AttrList =
@@ -1347,6 +1380,9 @@ static inline bool isPointerArithmeticInst(const llvm::Value *V,
     if (funcName == "julia.pointer_from_objref") {
       return true;
     }
+    if (funcName == "julia.gc_loaded") {
+      return true;
+    }
     if (funcName.contains("__enzyme_todense")) {
       return true;
     }
@@ -1405,6 +1441,10 @@ static inline llvm::Value *getBaseObject(llvm::Value *V,
         V = Call->getArgOperand(0);
         continue;
       }
+      if (funcName == "julia.gc_loaded") {
+        V = Call->getArgOperand(1);
+        continue;
+      }
       if (funcName == "jl_reshape_array" || funcName == "ijl_reshape_array") {
         V = Call->getArgOperand(1);
         continue;
@@ -1482,6 +1522,37 @@ static inline const llvm::Value *getBaseObject(const llvm::Value *V) {
   return getBaseObject(const_cast<llvm::Value *>(V));
 }
 
+static inline llvm::SetVector<llvm::Value *>
+getBaseObjects(llvm::Value *V, bool offsetAllowed = true) {
+  llvm::SmallPtrSet<llvm::Value *, 1> seen;
+  llvm::SetVector<llvm::Value *> results;
+  llvm::SmallVector<llvm::Value *, 1> todo = {V};
+
+  while (todo.size()) {
+    auto obj = todo.back();
+    todo.pop_back();
+    if (seen.contains(obj))
+      continue;
+    seen.insert(obj);
+
+    if (auto PN = llvm::dyn_cast<llvm::PHINode>(obj)) {
+      for (auto &x : PN->incoming_values()) {
+        todo.push_back(x);
+      }
+      continue;
+    }
+
+    auto cur = getBaseObject(obj, offsetAllowed);
+    if (cur != obj) {
+      todo.push_back(cur);
+      continue;
+    }
+
+    results.insert(obj);
+  }
+  return results;
+}
+
 static inline bool isReadOnly(const llvm::Function *F, ssize_t arg = -1) {
   if (F->onlyReadsMemory())
     return true;
@@ -1580,7 +1651,7 @@ static inline bool isNoCapture(const llvm::CallBase *call, size_t idx) {
     // may be nocapure/readonly, but the actual arg (which will be put in the
     // array) may not be.
     if (F->getCallingConv() == call->getCallingConv())
-      if (F->hasParamAttribute(idx, llvm::Attribute::NoCapture))
+      if (idx < F->arg_size() && F->getArg(idx)->hasNoCaptureAttr())
         return true;
     // if (F->getAttributes().hasParamAttribute(idx, "enzyme_NoCapture"))
     //   return true;
@@ -1678,6 +1749,11 @@ static inline bool isNoEscapingAllocation(const llvm::Function *F) {
   case Intrinsic::exp:
   case Intrinsic::cos:
   case Intrinsic::sin:
+#if LLVM_VERSION_MAJOR >= 19
+  case Intrinsic::tanh:
+  case Intrinsic::cosh:
+  case Intrinsic::sinh:
+#endif
   case Intrinsic::copysign:
   case Intrinsic::fabs:
     return true;
@@ -1934,7 +2010,11 @@ static inline llvm::Attribute::AttrKind PrimalParamAttrsToPreserve[] = {
     llvm::Attribute::AttrKind::NoFree,
     llvm::Attribute::AttrKind::Alignment,
     llvm::Attribute::AttrKind::StackAlignment,
+#if LLVM_VERSION_MAJOR >= 20
+    llvm::Attribute::AttrKind::Captures,
+#else
     llvm::Attribute::AttrKind::NoCapture,
+#endif
     llvm::Attribute::AttrKind::ReadNone};
 
 // Parameter attributes from the original function/call that
@@ -1949,7 +2029,11 @@ static inline llvm::Attribute::AttrKind ShadowParamAttrsToPreserve[] = {
     llvm::Attribute::AttrKind::NoFree,
     llvm::Attribute::AttrKind::Alignment,
     llvm::Attribute::AttrKind::StackAlignment,
+#if LLVM_VERSION_MAJOR >= 20
+    llvm::Attribute::AttrKind::Captures,
+#else
     llvm::Attribute::AttrKind::NoCapture,
+#endif
     llvm::Attribute::AttrKind::ReadNone,
 };
 #ifdef __clang__
@@ -1958,6 +2042,67 @@ static inline llvm::Attribute::AttrKind ShadowParamAttrsToPreserve[] = {
 #pragma GCC diagnostic pop
 #endif
 
+static inline llvm::Function *
+getIntrinsicDeclaration(llvm::Module *M, llvm::Intrinsic::ID id,
+                        llvm::ArrayRef<llvm::Type *> Tys = {}) {
+#if LLVM_VERSION_MAJOR >= 20
+  return llvm::Intrinsic::getOrInsertDeclaration(M, id, Tys);
+#else
+  return llvm::Intrinsic::getDeclaration(M, id, Tys);
+#endif
+}
+
+static inline llvm::Instruction *getFirstNonPHIOrDbg(llvm::BasicBlock *B) {
+#if LLVM_VERSION_MAJOR >= 20
+  return &*B->getFirstNonPHIOrDbg();
+#else
+  return B->getFirstNonPHIOrDbg();
+#endif
+}
+
+static inline llvm::Instruction *
+getFirstNonPHIOrDbgOrLifetime(llvm::BasicBlock *B) {
+#if LLVM_VERSION_MAJOR >= 20
+  return &*B->getFirstNonPHIOrDbgOrLifetime();
+#else
+  return B->getFirstNonPHIOrDbgOrLifetime();
+#endif
+}
+
+static inline void addCallSiteNoCapture(llvm::CallBase *call, size_t idx) {
+#if LLVM_VERSION_MAJOR >= 20
+  call->addParamAttr(
+      idx, llvm::Attribute::get(call->getContext(), llvm::Attribute::Captures,
+                                llvm::CaptureInfo::none().toIntValue()));
+#else
+  call->addParamAttr(idx, llvm::Attribute::NoCapture);
+#endif
+}
+
+static inline void addFunctionNoCapture(llvm::Function *call, size_t idx) {
+#if LLVM_VERSION_MAJOR >= 20
+  call->addParamAttr(
+      idx, llvm::Attribute::get(call->getContext(), llvm::Attribute::Captures,
+                                llvm::CaptureInfo::none().toIntValue()));
+#else
+  call->addParamAttr(idx, llvm::Attribute::NoCapture);
+#endif
+}
+
+[[nodiscard]] static inline llvm::AttributeList
+addFunctionNoCapture(llvm::LLVMContext &ctx, llvm::AttributeList list,
+                     size_t idx) {
+  unsigned idxs = {(unsigned)idx};
+#if LLVM_VERSION_MAJOR >= 20
+  return list.addParamAttribute(
+      ctx, idxs,
+      llvm::Attribute::get(ctx, llvm::Attribute::Captures,
+                           llvm::CaptureInfo::none().toIntValue()));
+#else
+  return list.addParamAttribute(ctx, idxs, llvm::Attribute::NoCapture);
+#endif
+}
+
 static inline llvm::Type *getSubType(llvm::Type *T) { return T; }
 
 template <typename Arg1, typename... Args>
@@ -1997,14 +2142,42 @@ static inline bool isSpecialPtr(llvm::Type *Ty) {
   return AddressSpace::FirstSpecial <= AS && AS <= AddressSpace::LastSpecial;
 }
 
+#if LLVM_VERSION_MAJOR >= 20
+bool collectOffset(
+    llvm::GEPOperator *gep, const llvm::DataLayout &DL, unsigned BitWidth,
+    llvm::SmallMapVector<llvm::Value *, llvm::APInt, 4> &VariableOffsets,
+    llvm::APInt &ConstantOffset);
+#else
 bool collectOffset(llvm::GEPOperator *gep, const llvm::DataLayout &DL,
                    unsigned BitWidth,
                    llvm::MapVector<llvm::Value *, llvm::APInt> &VariableOffsets,
                    llvm::APInt &ConstantOffset);
+#endif
 
 llvm::CallInst *createIntrinsicCall(llvm::IRBuilderBase &B,
                                     llvm::Intrinsic::ID ID, llvm::Type *RetTy,
                                     llvm::ArrayRef<llvm::Value *> Args,
                                     llvm::Instruction *FMFSource = nullptr,
                                     const llvm::Twine &Name = "");
+
+bool isNVLoad(const llvm::Value *V);
+
+//! Check if value if b captured after definition before executing inst.
+//! If checkLoadCaptured != 0, also consider catpures of any loads of the value
+//! as a capture (for the number of loads set).
+bool notCapturedBefore(llvm::Value *V, llvm::Instruction *inst,
+                       size_t checkLoadCaptured);
+
+// Return true if guaranteed not to alias
+// Return false if guaranteed to alias [with possible offset depending on flag].
+// Return {} if no information is given.
+#if LLVM_VERSION_MAJOR >= 16
+std::optional<bool>
+#else
+llvm::Optional<bool>
+#endif
+arePointersGuaranteedNoAlias(llvm::TargetLibraryInfo &TLI, llvm::AAResults &AA,
+                             llvm::LoopInfo &LI, llvm::Value *op0,
+                             llvm::Value *op1, bool offsetAllowed = false);
+
 #endif // ENZYME_UTILS_H
